#!/usr/bin/perl -w

#    crtvslcmd.pl -- create a visual command, generate a perl script
#    (C) 2010 Edgar <emax> Hermanns, emax AT berlios DOT de

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.


# todo: list attribute x,y,z
# todo: range attribut x:y
# todo composite list/range  x,y,a:c

=head1 NAME 

B<crtvslcmd.pl> -- create a curses-driven perl command

=head1 SYNOPSIS

B<crtvslcmd.pl> [-h|--help] [--lang=LANG] [--version|-v] --input INPUT --output OUTPUT

=head1 DESCRIPTION

B<crtvslcmd.pl> is a script which reads a textfile with a 'visual' description of a parameter entry 
screen from stdin and writes a perl program with an entry screen for those parameters to stdout.

The name 'B<crtvslcmd>' is a tribute to the B<AS/400> (aka System-i, i5) operating system B<OS400> 
and forms an abbreviation of the 'B<Create Visual Command>' phrase. Such generic abbreviations are 
used throughout the B<OS400> operating system and deliver in my eyes a genious nomenklature to the 
function provided by a particular command.

=head1 OPTIONS

=over

=item * B<-h|--help>  shows this screen

=item * B<LANG> sets the ouptut-language. Messages and screen-constants which are not user-defined will be created in the specified language. Languages currently supported are B<de> and B<en>.

=item * B<INPUT> is the file which contains the description of the entry screen.

=item * B<OUTPUT> is the file to which the resulting program is written

=over

=head1 PREREQUISITES

Perl must be installed, als well as the Perl modules Curses, Curses::UI and Getopt::Long.
If they are not already part of your Perl installation they can easily be installed with B<cpan>.
 
=head1 OPERATION

This script understands a syntax which provides a 'visual' description of a parameter set to be 
entered by a curses-interface. Each entry-field of this interface corresponds to one parameter. 

The parameters can optionally stored in an rc-file to provide the values which have been entered 
during the last call of the program as the default values for the next call. A function for
the parameter validation and for plausibility checks is generated. Display attributes and colors as
supported by the perl Curses:UI package can be applied. An number of automatic checks for the field 
contents can be used.

The generated script supports commandline parameters for each entry field. The syntax to provide an entry 
field with a value from the commandline is very simple: B<fieldname=value>

The description is read from INPUT, output is carried out to OUTPUT. 

B<CAUTION>: If the OUTPUT file already exists, it will be overwritten.

An additional file is created, which has the basename of the input file and the extension B<.pm>. It is stored in the same directory where the output file is stored. If the file already exists, it remains untouched. This file will be included from the generated script as a package and contains the B<Run> method, which is intented to do the final work. After initial creation, you can change the B<Run> method at will, it will not be changed a second time, so your individual code will never got lost. In the B<Run> method you have full access to all field-values and attributes You have as well access to the log-message which will be shown on the screen when the B>Run> method returns.

=head1 CONFIGURATION

There is currently nothing to configure.

=head1 INPUTFILE FORMAT

The input file is organized in three sections, each of which separated by a double percent sign similar to
a 'lex' file. The sections are used as follows:

=over

=item * B<Header section:> Defines general attributes of the script and the screen which will be generated.

=item * B<Visual section:> Defines the design of the screen and the entry fields in a 'visual' way.

=item * B<Attributes section> Defines extended attributes for the texts and entry fields of the screen.

=back

See the B<example *vsl files> supplied with this script.

=head2 Inputfile Header Section

The header section contains keyword/value definitions, one pair per line. 

A line starting with a '#' is considered a comment. Empty lines are ignored. Any other line is considered a 
definition, where the first non-whitespace expression is considered the keyword which is terminated by 
whitespace. Everthing after this whitespace is considered the value for the keyword in question. Keywords 
are not case sensitive. and may thus be written in upper- or lowercase.

Currently the following keywords are implemented (for better readability they are written in uppercase):

=over

=item * TITLE - A string which will be printed in the very first line of the screen in reverse font. The string may 
contain whitespaces. The TITLE is intended to provide a short prompt of what the script does.

=item * RCFILE - A file name where the default values for the generated script are stored. The file must be a valid 
filename for the operation system in question and may contain globbing wildcard characters. The filename 
shall not contain whitespaces.

=item * MAXBLANKS - This integer value determines, how much blanks may subsequently be embedded in a constant. If more subsequent blanks are found in a constant, a new constant is generated. By default no such separation is made.

=item * ORDER - Defines the general order in which entry fields are processed. This feature can be helpful if you want to process a screen in column-priority. The possible values are as follows:

=over

=item * B<y,x> The topmost line is processed first, with the cursor advancing through all fields in that line from left to right. Then the next line below is processed in the same way until the entire screen has been processed. This is the default behaviour.

=item * B<y,-x> The same as B<y,x>, except that the field processing occurs from the right to the left for each line.

=item * B<-y,x> The same as B<y,x>, except that processing goes bottom up, i.e. starts in the lowermost line and ends in the topmost line. 

=item * B<-y,-x> The same as B<-y,x>, except that the field processing occurs from the right to the left for each line.

=item * B<x,y> The leftmost field is processed first, no matter in which line it is. Then, the next field which is next in terms of its horizontal position, is processed. If two or more fields share the same horizontal position, they are processed from top to bottom.

=item * B<x,-y> The same as B<x,y>, except that lines are processed from bottom to top in case that one or more fields share the same horizontal position.

=item * B<-x,y> The same as B<x,y>, except that processing starts with the rightmost field and goes to the left.

=item * B<-x,-y> The same as B<-x,y>, except that processing except that lines are processed from bottom to top in case that one or more fields share the same horizontal position.

=back

B<NOTE:> The general order of field-processing as described above can be overridden for individual fields (see B<Attributes Section>). 

=item * AUTHOR - An optional author information which will be printed on top of the script by means of a comment. If you use newlines B<'\n'>, the string must be double-quoted, and each new line must be preceeded by a hash B<'#'>.

=item * COPYRIGHT - An optional copyright information which will be printed on top of the script by means of a comment. If you use newlines B<'\n'>, the string must be double-quoted, and each new line must be preceeded by a hash B<'#'>.

=item * VERSION - An optional version information which will be printed on top of the script by means of a comment. If you use newlines B<'\n'>, the string must be double-quoted, and each new line must be preceeded by a hash B<'#'>.

=item * DESCRIPTION - An optional description which will be printed on top of the script by means of a comment. If you use newlines B<'\n'>, the string must be double-quoted, and each new line must be preceeded by a hash B<'#'>.

=back

As all keywords are stored internally in terms of a hash, unknown keywords will not lead to an error, 
they will however not have any impact on the behaviour of the script.

=head2 Inputfile Visual Section

The screen definition section is separated from the header section by a double percent-sign '%%' in 
the very first column. This section is defined in 'visual' mode and resembles very closely how the 
final screen on the monitor shall look like when the new script is called. 

You can easily 'paint' the design of the screen by simply typing the text-constants in the exact lines and 
positions where they shall appear on the screen. To define an entry field, us the syntax described below. 
If a constant or an entry field on the screen is supposed to be display whith special attributes, e.g. 
a particular color, you can define this in the third section of the input file.

=head2 Defining Entry Fields

If you want to define an entry field on the screen, simply indicate this with a leading dot '.', immediately 
followed by a name. The syntax of the name must comply to perl variable names, not including the '$' sign.

B<Example>: The following screen definition defines a screen with the Title 'Example', a constant which reads 
'Customer number:" and a customer number entry field:

 TITLE Example
 %%
 Customer number:.custno
 %%
 
The following rules apply for the entry field 'custno' as defined above:

=over

=item * The entry field on the resulting screen will appear at the position, where the word 'custno' begins: 
The leading dot is ignored and will be replaced by a blank. 

=item * The field will internally have the name 'custno'. 

=item * The length of the entry-field is 6 bytes. The six bytes correspond to the length of the variable name 
'custno'. If you want the entry field to be longer than its variable name, append underscores until the 
intended length of the entry-field is matched. The underscores will not be part of the field name.

=item * If the variable name ends with a dollar '$' sign, the length of the field is expanded as much as possible,
i.e: until the next constant, entry-field or up to the end of the line.

=item * If a variable name is terminated with a dollar '$' sign, the contents of the entry field scroll to the 
left if the cursor reaches the right end of the field during entry. So the entry field length is in fact infinite.

=item * All extended definitions for this field in section three must be made with the prefix 'custno' (explained 
in detail in the B<Inputfile Attributes Section> chapter below).

=item * A variable name must not be used more than once in the visual section.

=item * As of the current release, all constant texts must start with an alpha sign.

=back

=head2 Inputfile Attributes Section

The attribute section is separated from the visual section by a double percent-sign 'B<%%>' in the very first column. 

In this section you can define additional attributes for each entity which has been used in the visual section. 
This includes the entry fields and the constants as well.

A line starting with a '#' is considered a comment. Empty lines are ignored. Any other line is considered an 
attribute definition.

To define an attribute for an entry field, you have to write the name of the entry field from the visual section first, followed by a dot '.' and appended by the attribute-name. A whitespace must follow this expression, which in turn is followed by the attribute value. The attribute value may contain whitespaces

B<Example>: define 'prompt' text for the field 'cusno':

 cusno.prompt       Please enter the customer number

As of this release, the following attributes for entry-fields are supported (listed in alphabetical order):

=over

=item * B<composite>: The entry field will unconditionally be filled with the components defined here. Components can be constants and variables which are defined somewhere on the screen. Variables have to be defined in the form B<${variable}>. 

B<Example>: Two fields 'dir' and 'file' are defined in the screen. You have another field named 'path' which shall contain the composite of boith, dir and path. You can define this as follows:

 path.composite   ${dir}/${file}

This defintion uses two variables and the constant slash in between. The contents of the path field are unconditionally replaced by the aggregate of the above. Unconditionally means that any content of the field will unconditionally be overwritten. If you want the composite value to be created only in case of an empty field, use the B<default attribute> instead. It is capable to process composite values as well.

B<NOTE>: It makes sense, to make a B<composite> field B<non-focusable> as well, as any entry in such a field will be discarded anyway. To make a field B<non-focusable> use the B<-focusable> attribute described below.

=item * B<default>: If a field is empty after trimming, it will be set to this value. For the value, the same syntax applies as described for the B<composite attribute>.

=item * B<dir>: Checks the field contents against one or more directories. This attribute accepts one or two numerical values, separated by a colon. The following rules apply:

=item * B<help>: A 'help' attribute provides a longer help text which is available for the particular field by pressing the F1-key. The text ist displayed in a pop-up window and can have multiple lines. If you want to use newline characters in the text, you have to use escape sequences '\n' and to to enclose the text in double-quotes.

=over

=item * B<0> There must not be a file of the given name. No name globbing is applied.

=item * B<1> Any value other then zero means, that there must be one file of the given name. No name globbing is applied.

=item * B<n:m> There must exactly be B<n> to B<m> files of the given pattern. Name globbing is applied. If B<n> is omitted, zero is assumed. If B<m> is omitted, an infinite number is assumed. At least one of B<m> or B<n> must be defined for globbing activation, The colon B<:> is mandatory.

=back

B<Examples>: With B<3:8> there must be 3 to 4 files which match the given pattern. With B<:2> there must be not more then 2 files which match the given pattern. With B<2:> there must be at least 2 files which match the given pattern. With B<0:0> there must not be any file which matches the given pattern. 

The latter example differs from a single zero (i.e. B<not using a colon>) in so far, that a single zero does not cause globbing to be applied, thus treating the field contents like a name and not like a pattern.

=item * B<mandatory>: If set to a non-zero value, checks against empty fields. Optional whitespace removal is done prior to this check, thus causing a whitepsace-only content not to be accepted.

=item * B<noread>: If set to anything other than 0 and rc-file processing is defined, the value of entry field will not be read from the rc-file. You can use this attribute to make a field always having an empty value whenever the script is started. If rc-file processing is defined, the field will anyway be written to the rc-file. This can be helpful for documentation purposes or for other scripts which use the same rc-file as well.

=item * B<nowrite>: If set to anything other than 0 and rc-file processing is defined, the value of entry field will not be written to the rc-file. You can use this attribute to make a field always having the same preset value when the script is started. This can be achieved by writing the value in question manually to the rc-file. Unless you did not specify 'noread', the field will always be read from the file on start-up, thus showing the valued stored there whenever the script is started.

=item * B<prompt>: A 'prompt' attribute is a text, which appears on the bottom line of the srceen, when the cursor is positioned in the corresponding field. It is intended to give a short help for what has to be entered in that field.

=item * B<rename>: With this attribute, a field can be assigned a different name. This name will be used internally throughout the generated script instead of the original name which was used in the visual section. This attribute is provided to assign a more meaningfull name to very short fields from the visual section.

=item * B<wscompress>: If set to a non-zero value, compresses embedded whitespace to one blank each.

=item * B<wsdelete>: If set to a non-zero value, removes all whitespace from the field.

=item * B<wstriml>: If set to a non-zero value, deletes left hand whitespace from the field.

=item * B<wstrimr>: If set to a non-zero value, deletes right hand whitespace from the field.

=back

You can also add attribues to the text constants which have been used in the visual section. As the text constants do not have 'field-names', they are assigned generic names during script generation. The generic names are made of the constant text itself, whereupon all non-word characters, i.e. everthing which is not an alpha character or an underscore, is replaced by an underscore '_'. In case of the "Customer number:" constant, the label "Customer_Number_" will be assigned internally. 

An attribute is assiged in the same way as described for entry fields, except that the token B<Label:> has to be used as a prefix for the label name. 

B<How to achieve a list of all label names of this kind is described in the B<'UTILZATION AND EXMAPLES'> chapter>. 

The following definition assigns a blue background color to the "Customer Number:" constant:

 Label:Customer_Number_.-bg  blue

All attributes starting with a dash '-' are handed over to Curses::UI, regardless of the entity-type: entry-fields or labels. In general, all attributes supported by the Curses::UI interface may be used in the attributes section. See the B<Curses::UI documentation and sources> for details. In particular, the TextEditor.pm source file gives a comprehensive overview of the attributes provided by Curses::UI. 

As of this writing, the following Curses::UI attributes have been testet and been proven to work as expected:

=over

=item * B<-bg>: The -bg attribute determines the background color of the entry field.

=item * B<-focusable>: If set to zero, causes the field to be read-only, so that the field cannot be edited. It will not be underlined.

=item * B<-password>: The password attribute sets the character which is displayed instead of the character which had been entered into the field.

=item * B<-regexp>: The B<-regexp> attributes causes the Curses::UI package to check the contents of a field against the rule expressed by the regular expression. You can not manually enter a value which does not macht the regular expression.

It can not be guaranteed, that all Curses::UI attribute utilization will work as expected.

=back

See also the B<example .vsl files> supplied with this script.

=head1 Internal Script Organization

The generated script will be curses-driven and behaves as follows:

=over

=item * If an rc-file was defined in the header section, the values from the rc-file are read

=item * The commandline parameters are read and assigned to the entry-fields, thus overwriting the values
found in the rc-file.

=item * The dialog-window is created and displayed, with the entry fields showing the values found in 
the rc-file or provided on the commandline respectively.

=item * If the user presses the 'F3'-key, the script ends with no action and returns control to the caller. 
In that case, the return code is '3'.

=item * The values can be entered. For each field which is entered by the cursor, the 'prompt' value which 
was assigned for this field in the attribute section is displayed at the bottom line of the screen. If no such 
attribute was assigned, the name of the field is displayed instead.

=item * If the user presses the 'F1' key, the 'help' value which was assigned for this field in the attribute 
section is displayed in a pop-up window. If no such attribute was assigned, the name of the field is displayed 
together with a the information, that there is ho help available for the field in question.

=item * If a B<-regexp> attribute has been defined for the field, you can not edit the field in a way that contradict to the regular expression.

=item * If the user presses the 'F5'-key, all fields are retained the values they had when the script was 
initially started, thus discarding any changes made by the user.

=item * If an rc-file was specified in the header-section and the user presses the 'F10'-key, all values which 
are currently in the entry-fields are saved to the rc-file, B<even if they are not valid>.

=item * If the user presses the ENTER key, the predefined B<CheckScreen()> method is executed. 

=item * In the B<CheckScreen> method, all field content is checked generically as defined in the B<Attributes Section>. If a field does not meet the required conditions, an error message is displayed on a pop-up screen, and control is returned to the screen with the cursor placed in the entry-field in question.

=item * If all fields have passed the checking procedure, the B<Run> method from the package which were automatically been created for this script is called. 

=item * When the B<Run> method return, a logging window is popped up to display the messages from the B<Run> method.

=item * If the B<Run> method returned the Value B<1>, all values from the entry fields are saved to the rc-file, the dialog is destroyed and the program ends. 

=item * If the return code is anything other than '1', control is returned to the screen.

Remember, that the return code '3' is already in use for a termination triggered by the 'F3-key.

=back

Inside the generated script you will find a number of comments which tell you what to do to make the script do 
what it's supposed to do. 

B<NOTE:> The script is designed so that you can do everything in the B<Run> method, thus letting the generated script be completely untouched. That way, you can change the design and add/remove parameters at will, without having to write everything again after script generation.

=head1 UTILIZATION AND EXAMPLES

One or more exmaple-files are supplied together with this srcipt. Have a look at these examples, inside 
you will find a number of use-cases for the features described above.

To generate a perl-script with one of the example-files, do the following:

  me@mymachine:~$ crtvslcmd.pl --input example1.vsl --output example1.pl
  me@mymachine:~$ chmod ug+x example1.pl
  me@mymachine:~$ ./example1.pl

The first statement calls this script, reads the 'example1.vsl' file from stdin and writes the result to
a file named 'example1.pl'. With the second line, you make the 'example1.pl' file executable. With the third
command, you call the newly generated program.

If you call the 'example1.pl' another time, the values which have been entered upon the last succesfull call 
of the script will be shown in the entry-fields. They have been stored in the '~/.example1rc' file. To 
display the contents of this file, do the following:

  me@mymachine:~$ cat ~/.example1rc

To overwrite a value from the rc-file, call the script with commandline parameters to overwrite the 
value(s) in question:

  me@mymachine:~$ ./example1.pl parm=value

where 'parm' is the field name in question, whilst 'value' is the value which you want to assign to the entry-field.
The field names are displayed in the pop-up window which you can get by pressing the 'F1'-key when the cursor is in 
the corresponding entry-field.

If you want to achieve a list of all constant-label names, call the script like this:

  me@mymachine:~$ crtvslcmd.pl --debug=:DCN: --input example1.vsl --output  example1.pl

All constant names which can be used as 'Label'-identifiers within the attributes sections are displayed together
with their position on the screen. The line and column numbers which are displayed refer to the area inside
the surrounding border of the screen. This area starts at the upper left corner within the border, counting 
lines and columns from zero upwards and thus starting with the coordinates 0,0.

=head1 Restrictions

All documentation refers to the utilization of this software under linux or any ..ix operation system. For other 
platforms, the instructions provided above can be applied accordingly.

=head1 Conclusion

This utility is a result of a 'quick & dirty' scripting activity, thus not providing an elegant architecture. But the 
results are pretty good usable and provide a robust and smart interface for simple daily tasks user might have to
do in their daily business.

If you find bugs or have suggestions for improvements, don't hesitate to contact me at

 emax At berlios Dot de

=head1 AUTHOR

Edgar (emax) Hermanns 

=cut

use strict;
use Getopt::Long;
use File::Basename;
use Text::Wrap;
$Text::Wrap::columns = 78;         # Wrap at 
$Text::Wrap::huge    = 'overflow'; # or 'die' or 'wrap'

my $dbg = ""; #"ALL:"; #:WEN:RB:RP:SV:";
my $startx = 1;
my $starty = 1;
my %prolog =();
my %yxParts = ();
my %ordYxParts =();      # pointer to yxParts in order of field processing
my @vars = ();
my %fieldAttribs = ();
my %labelAttribs = ();
my $firstVar;
my $needsCheckFD = 0;
my $ypos = $starty-1;
my $maxEmbeddedBlanks = 2;
my $lang = 'de';

my $helpme;
my $version;
my $input = "";
my $output = "";
my $vslname;
my $vslpath;

my $vsn = "1.2";
my $inp;
my $out;

my %texts = (
    de  => { FIELD_AND_PATTERN => 'Field [$varname]\n\nPattern: [$pattern]\n\n',
             FIELD             => 'Feld',
             FIELDNAME         => 'Feldname',
             NOHELP            => 'Fuer dieses Feld ist leider keine Hilfe verfuegbar',
             HELP              => 'Hilfe',
             UNKNOWNPFMT       => 'unbekanntes Parameter format',
             INVALIDPARM       => 'ungültiger Parameter',
             COULDNTREADRC     => 'konnte die rc-Datei RCFILE nicht lesen',
             COULDNTOPENRC     => 'konnte die rc-Datei RCFILE nicht zum Schreiben öffnen',
             VALUESSAVED       => 'Werte wurden gesichert',
             EXISTMUSTNOT      => '${type} existiert bereits (nicht erlaubt).',
             NOTEXISTHASTO     => '${type} existiert nicht (muss vorhanden sein).',
             FATALCHECKCOND    => 'FATAL: unerwartete ${type} Prüfbedingung.',
             NOSUCHFILE        => 'Es gibt kein ${type} dieses Names. Mindestens $min müssen/muss vorhanden sein.',
             TYPEEXISTS        => 'Ein/e solche/s ${type} existiert bereits (nicht erlaubt).',
             COUNTTYPEEXIST    => '$count ${type}/s dieses Namens exisieren bereits. (Null sind erlaubt).',
             MORETHANEXIST     => 'Es existieren bereits mehr als $max ${type}/s dieses Names [$count wurden gefunden].',
             LESSTHANEXIST     => 'Es exisitieren weniger als $min ${type}/s dieses Namens [$count wurden gefunden].',
             VALUESENTERED     => 'Diese Werte wurden eingegeben:\n\n',
             PARMSNVALUES      => 'Parameter                       Wert\n',
             PARMSNVALUESUL    => '-----------------------------   --------------------------------\n',
             EXAMPLEMSG        => 'Die nächste Anzeige dient nur als Beispiel, man kann eigene Parameter Prüfungen einbauen. Aktuell werden alle Felder gegen die Bedingungen geprüft, die im Attribute-Abschnitt definiert wurden.\n\nBeachten Sie dazu bitte auch \'CheckScreen()\' Methode in diesem Script.',
             PROCINPROGRESS    => 'Verarbeitung läuft ... \n----------------------\n',
             DONE              => '\nOk, fertig.',
             MANDATORY         => 'Pflichtfeld: [$name]\n\nBitte geben Sie einen Wert ein.',
             LOG               => 'Protokoll',
             BOTTOMLINE1       => 'F1: hilfe  F3: abbrechen  F5: reset  ',
             BOTTOMLINE2       => '   ENTER: execute         ',
             F10               => "F10: Werte sichern",
    },
    en  => { FIELD_AND_PATTERN => 'Field [$varname]\n\nPattern: [$pattern]\n\n',
             FIELD             => 'Field',
             FIELDNAME         => 'Fieldname',
             NOHELP            => 'Sorry, for this field is no help available.',
             HELP              => 'Help',
             UNKNOWNPFMT       => 'unknown parameter format',
             INVALIDPARM       => 'invalid parameter',
             COULDNTREADRC     => 'couldn\'t read rc-file RCFILE',
             COULDNTOPENRC     => 'couldn\'t open rc-file RCFILE for write',
             VALUESSAVED       => 'Values were saved',
             EXISTMUSTNOT      => '${type} exists (but must not)',
             NOTEXISTHASTO     => '${type} doesn\'t exist (but has to)',
             FATALCHECKCOND    => 'FATAL: unknown ${type} checking condition',
             NOSUCHFILE        => 'no such ${type}s exist (but at least $min ${type}s have to)',
             TYPEEXISTS        => 'such a ${type} already exists (but must not)',
             COUNTTYPEEXIST    => '$count such ${type}s already exist (but must not)',
             MORETHANEXIST     => 'more than $max such ${type}s already exist [found $count matches].',
             LESSTHANEXIST     => 'less then $min such ${type}s currently exist [found $count matches].',
             VALUESENTERED     => 'The values you have entered:\n\n',
             PARMSNVALUES      => 'Parameter                       value\n',
             PARMSNVALUESUL    => '-----------------------------   --------------------------------\n',
             EXAMPLEMSG        => 'The next window is only an example, you can fill in your own parameter checking. Currently, the fields are checked against all conditions which were defined in the attributes section.\n\nSee the \'CheckScreen()\' method in this script',
             PROCINPROGRESS    => 'processing is in progress ...\n-----------------------------\n',
             DONE              => '\nOk, done.',
             MANDATORY         => 'Mandatory field: [$name]\n\nPlease enter a value.',
             LOG               => 'Log',
             BOTTOMLINE1       => 'F1: help   F3: discard   F5: reset   ',
             BOTTOMLINE2       => '   ENTER: ausführen       ',
             F10               => "F10: save values"
    }
);

#==================== CODE FRAGMENTS ====================
my %code = (
# ---------- intro1 ----------
    intro1 => q
|#!/usr/bin/perl -w
#------------------------------------------------------------------------------
#                          This script was generated by
#                                 'crtvslcmd.pl'
#                                     the
#                          Perl Visual Command Creator.
#
#                                 An emax tool.
#------------------------------------------------------------------------------
|,

# ---------- header ----------
    header => q
|# type   value
|,

# ---------- intro2 ----------
    intro2 => q
|
# push the path were the script is to INC
BEGIN { $0 =~ s!^(.*/)[^/]*$!$1!; push (@INC, $1); }

use strict;
use Curses;
use Curses::UI;
use vars qw( $_cui
             $_mainWin
             $_cols
             $_lines
             $_curField
             $_infoMsg
             $_promptFld
             $_firstField );
use PACKAGE;

my %_fields = ( 
|,

# ---------- field_decl ----------
    field_decl => q
|    fieldname => { |,

# ---------- field_attr ----------
    field_attr => q
|,@@attr=>attrval|,

# ---------- field_decl_close ----------
    field_decl_close => q
| },
|,

# ---------- fields_hash_close ----------
    fields_hash_close => q
|    ); # %_fields hash

my $_confirm = 0;
|,

# ---------- rcfile_decl ----------
    rcfile_decl => q
|my $_rcFile =glob('rcfile');
|,

# ---------- helper_funcs ---------- 
    helper_funcs => q 
|
sub getComposite()
{
    my $pattern = shift;
    my $result = "";
    while (length($pattern))
    {
        ($pattern =~ s/^\${(\w+)}//x) and $result .= $_fields{$1}->{value} and next;
        ($pattern =~ s/^([^\$]+)//x)  and $result .= $1 and next;
        $result .= $pattern; #unknown syntax, just append
        $pattern ="";
    }
    return $result;
}

sub onBlurAnyField () { $_curField = shift; } # onBlurAnyField ()
sub onFocusAnyField ()
{
    $_curField = shift;
    my $field = $_fields{$_curField->{name}};
    if ($_infoMsg)
    {
        $_promptFld->text($_infoMsg);
        $_infoMsg = undef;
    }
    elsif (exists $field->{prompt}) { $_promptFld->text($field->{prompt}); }
    else  { $_promptFld->text("FIELD: [".$_curField->{name}.']'); }
} # onFocusAnyField ()

sub Help()
{
    my $help = "FIELDNAME: [".$_curField->{name}."]\n\n";
    my $field = $_fields{$_curField->{name}};
    if (exists $field->{help}) { $help = $help.$field->{help}; }
    else { $help=$help."NOHELP"; }
    $_cui->dialog(-width=>$_cols-4, -title=>"HELP", -message=>$help, -focusable=>1 );
} # Help()

sub checkParms()     # write all parms in the form of parm='value' to a hash
{
    foreach my $arg( @ARGV )
    {
        ($arg =~ /^\W*(\w+)\s*=\s*(.+)\s*$/) or die "\n$0: UNKNOWNPFMT: $arg\n";
        my $fieldName = $1;
        my $value = $2;
        (!exists $_fields{$fieldName}) and die "\n$0 INVALIDPARM: [$fieldName]\n";
        $_fields{$fieldName}->{value} = $value;
    }
} # checkParms()

sub toScreen()
{
    foreach my $fieldName(keys %_fields) 
    { 
        $_mainWin->getobj($fieldName)->text($_fields{$fieldName}->{value});
    }
} # toScreen()

sub Reset() 
{ 
    foreach (keys %_fields) { $_fields{$_}->{value} = ""; }
    &readRcfile(); 
    &toScreen(); 
    $_firstField->focus();
} # Reset()
|,

# ---------- rcfile_rw ----------
    rcfile_rw => q
|
sub readRcfile()
{
    (! -f $_rcFile) and return;
    open (IN, "<$_rcFile") or die "\n COULDNTREADRC: $!";
    while (<IN>)
    {
        /^\#/ and next;
        /^\s*$/ and next;
        if (/^\s*(\w+)\s+(.*)/)
        {
            (!$_fields{$1}) and next;
            (exists $_fields{$1}->{noread} && $_fields{$1}->{noread} ne 0) and next;
            $_fields{$1}->{value}=$2;
        }
    }
    close IN;
} # readRcfile()

sub WriteRcfile()
{
    my %saved = ();
    # before rewriting (/and this deleting) the file, read all current content
    while (1) # dummy loop
    {
        (! -f $_rcFile) and last;
        open (IN, "<$_rcFile") or last;
        while (<IN>)
        {
            /^\#/ and next;     # skip comments
            /^\s*$/ and next;   # skip empty lines
            (/^\s*(\S+)\s+(.*)/) and $saved{$1} = $2;
        }
        close IN;
        last;
    } # end dummy loop

    # now write all our stuff
    open (OUT, ">$_rcFile") or  die "\n COULDNTOPENRC: $!\n";
    my $now = localtime;
    print OUT "# rcfile for $0, written $now\n";

    foreach my $fieldName (sort keys %_fields)
    {
        (!$_fields{$fieldName}) and next;       
        (exists $_fields{$fieldName}->{nowrite} && $_fields{$fieldName}->{nowrite} ne 0) and next;
        print OUT "$fieldName ", $_fields{$fieldName}->{value}, "\n";
        (exists $saved{$fieldName}) and delete $saved{$fieldName};
    }   

    # now, write all values which wre saved but not rewritten
    print OUT "\n# untouched values from $0, written $now\n";
    foreach my $fieldName (sort keys %saved)
    {
        print OUT "$fieldName $saved{$fieldName}\n";
    }
    close OUT;
    $_promptFld->text("VALUESSAVED.");
    $_curField->focus; # restore cursor
} # WriteRcfile()
|,

# ---------- mkdialog_intro ----------
    mkdialog_intro => q
|
sub mkdialog()
{
    $_cui      = new Curses::UI ( -color_support => 1, -clear_on_exit => 1 );
    $_lines    = Curses::UI::LINES();
    $_cols     = Curses::UI::COLS();
    $_curField = "";
    $_mainWin  = $_cui->add('mainWindow', 'Window', -height=>$_lines, 
                           -width=>$_cols, -border => 1, -centered => 1,
                           -title=>"ptitle");
|,

# ---------- mkdialog_bindings ----------
    mkdialog_bindings => q
|
    $_curField =  $_firstField = $_mainWin->getobj('firstvar');
    $_promptFld = $_mainWin->getobj('_prompt');
    $_mainWin->set_binding((\&CheckScreen), KEY_ENTER());
    $_mainWin->set_binding((\&Help), KEY_F(1));
    $_mainWin->set_binding((\&Quit), KEY_F(3));
    $_mainWin->set_binding((\&Reset), KEY_F(5));
|,

# ---------- key_f10_binding
    key_f10_binding => q
|    $_mainWin->set_binding((\&WriteRcfile), KEY_F(10));
|,

# ---------- on_funcs ----------
    on_funcs => q
|sub on_.fname.Focus  { my $this = shift; &onFocusAnyField($this); }
sub on_.fname.Change { my $this = shift; $_fields{hname}->{value} = $this->get(); }
sub on_.fname.Blur   { my $this = shift; &onBlurAnyField($this); }
|,

# ---------- on_funcs_end
    on_funcs_end => q
|#----------------------------------- end of "on" functions"
|,

# ---------- quit_func
    quit_func => q
|
sub Quit() { exit 3; } # Quit()
|,

# ---------- checkfd_func ---------- (different delimter required)
    checkfd_func => q
/
sub checkFD() # check file or directory
{
    # this function checks against file and directory names.
    # parameters: type, varname, glob, min, max, logtext
    my ($type, $varname, $glob, $min, $max, $logText) = @_;
    my $pattern = $_fields{$varname}->{value};
    my $count = 0;
    my @found = ();

    if ($glob)
    {
        while(glob($pattern))
        {
            if ($type eq 'file' && -f $_ || $type eq 'dir' && -d $_)
            {
                push (@found, $_);
                ++$count;
            }       
            # save found files to $_fields
            $_fields{$varname}->{globArr} = [@found];
        } # while glob(pattern)
    } else
    {
        ($type eq 'file' && -f $pattern || $type eq 'dir' && -d $pattern) and $count = 1;
    } 

    if ($count >= $min)
    {
        ($max < 0 || $count <= $max) and return 1;
    }

    $$logText = "FIELD_AND_PATTERN";
    if (!$glob)
    {
        $$logText .= 
            ($count == 1) ? "EXISTMUSTNOT" :
            ($count == 0) ? "NOTEXISTHASTO" :
            "FATALCHECKCOND"; 
    } else
    {
        $$logText .= 
            ($count == 0 && $min > 0)      ? "NOSUCHFILE" :
            ($count == 1 && $max == 0)     ? "TYPEEXISTS" :
            ($count > 1  && $max == 0)     ? "COUNTTYPEEXIST" :
            ($count > 1  && $max < $count) ? "MORETHANEXIST" :
            ($count > 1  && $min > $count) ? "LESSTHANEXIST" : 
            "FATALCHECKCOND"; 
    }
    $_mainWin->getobj($varname)->focus();
    return 0;    
} # checkFD
/,

# ---------- check_screen_intro1 ----------
    check_screen_intro1 => q
|
sub CheckScreen() {
    my $done = 0;
    my $logText = "?";
    $_confirm = 0;
|,

# ---------- check_screen_intro3 ----------
    check_screen_intro3 => q
|
    #----------------------------------------------------------------
    # add in your own parameter checking below. Currently, the checks
    # which have been defined in the attributes section are made.
    #----------------------------------------------------------------
    while(1) # a dummy loop so that we can use the 'last' keyword to jump to the end
    {
|,

# ---------- close_function
    close_function => q
|} # function()

|,

# ---------- dofieldcheck ----------
    dofieldcheck=> q
|        (!&checkField('fieldname', \$logText)) and last;
|,

# ---------- runepilog1 ----------
    runepilog1=> q
|
        $done = 1;
        last;
    } # end of while(!$done) dummy loop for parameter checking

    if (!$done) # an error occurred
    {
        # display logText if there
        $_cui->dialog(-width=>$_cols-4, -title=>" Error ", -focusable=>1, -message=>$logText);
        return;
    } 
    
    # everything was alright, but first confirm. Put focus on 1st field.
    ($_confirm) and $_firstField->focus() and return; # redisplay for confirmation before processing

    # processing will follow
    # in case that the script will cause uncaptured shell-output, you should leave curses temporarily
    $_cui->leave_curses();

    $logText = "PROCINPROGRESS";

    # implement your processing here
    # you can log information to $logText at will, it will be shown in the log-window
    my $Rc = Run(\$logText, \%_fields);

    # in case you left curses, restore its state after execution
    $_cui->reset_curses();

    ($Rc == 1) and $logText .= "DONE";
    $_cui->dialog(-width=>$_cols-4, -title=>" LOG ", -focusable=>1, -message=>$logText);
    ($Rc != 1) and return;
|,

# ---------- writercfile_call ----------
    writercfile_call=> q
|
    &WriteRcfile(); 
|,

# ---------- runepilog2 ----------
    runepilog2=> q
|
    $_cui->leave_curses();
    exit (0);

} # CheckScreen()

|,
# ---------- readrcfile_call ----------
    readrcfile_call=> q
|&readRcfile();              # read rcfile
|,

# ---------- final ----------
    final=> q
|&checkParms();              # read all parameters
&mkdialog();                # create the dialog

$_firstField->focus();      # start with first field
$_mainWin->modalfocus();    # run mainloop
$_mainWin->delete();        # cleanup
|,

# ---------- check_field ----------
    check_field => q
|
sub checkField()
{
    my ($name, $message) = @_;
    my $field = $_fields{$name};
    my $oldval = $field->{value};

    my $updfield = 0;
    (exists $field->{composite} && length($field->{composite}) != 0) and 
        $field->{value} = &getComposite($field->{composite});

    (exists $field->{wstriml} && $field->{wstriml} ne 0) and 
        $field->{value} =~ s/^\s+//g; # strip left hand ws
    
    (exists $field->{wstrimr} && $field->{wstrimr} ne 0) and
        $field->{value} =~ s/\s+$//g;    # strip right ws 

    (exists $field->{wscompress} && $field->{wscompress} ne 0) and
        $field->{value} =~ s/\s{2,}/ /g; # compress embedded ws

    (exists $field->{wsdelete} && $field->{wsdelete} ne 0) and
        $field->{value} =~ s/\s+//g;     # delete all ws
    
    (exists $field->{default} &&         # check for default 
     !length($field->{value})) and
     $field->{value} = &getComposite($field->{default});

    if ($field->{value} ne $oldval)
    {
        $_mainWin->getobj($name)->text($field->{value});
        $_confirm = 1;
    }

    if (exists $field->{mandatory} && $field->{mandatory} ne 0 && 
        length($field->{value}) == 0) # check mandatory
    {
        $$message = "MANDATORY";
        $_mainWin->getobj($name)->focus();
        return 0;
    } # end mandatory check

    if (exists $field->{fdcheck})
    { 
        my $checktype = $field->{fdcheck};
        (!&checkFD($checktype, $name, $field->{glob}, $field->{min}, $field->{max}, $message)) and return 0;
    }
    
    return 1;
} # end checkField

|,


# ---------- pkg ----------
    pkg=> q
|# this package was automatically created
# by crtvslcmd.pl, the visual command creator.
# the package name is 'PKG'
# fill in this script as needed. 
# It will not be overwritten.

sub Run
{
    my ($log, $fields) = @_;

    # you can completely delete this code until the '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>' marker
    $logText  = "VALUESENTERED";
    $logText .= "PARMSNVALUES";
    $logText .= "PARMSNVALUESUL";

    foreach (sort keys %$fields)
    {
        (/^_/) and next;
        $$log .= sprintf "%-32s[%s]\n", $_, $$fields{$_}->{value}, "\n";
    }    
    # '>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>' marker
    return 1; # the return is optional
}
1;
|,

# ---------- empty ----------
    empty=> q
|            # this is an empty fragment
|,

);



#================ END OF CODE FRAGMENTS =================

sub getText()
{
    # this function reads a text in $lang from %texts and replaces all
    # words given by the keys from the repl-hash with their key-values
    my ($textKey, %repl) = @_;

    (!exists $texts{$lang}->{$textKey}) and die ("\n\nERROR: text key [$textKey] does not exist in language [$lang]\n\n");

    my $text = $texts{$lang}->{$textKey};
    $text =~ s/\b$_\b/$repl{$_}/g  foreach (keys %repl);
    return  $text;
} # getFragment

sub getFragment()
{
    # this function reads a code fragment from %code and replaces all
    # words given by the keys from the repl-hash with their key-values
    my ($fragName, %repl) = @_;
    (!exists $code{$fragName}) and die ("\n\nERROR: Code fragment [$fragName] does not exist\n\n");

    my $frag = $code{$fragName};
    $frag =~ s/\b$_\b/$repl{$_}/g  foreach (keys %repl);
    return  $frag;
} # getFragment

sub setDefaults()
{
    $prolog{title} = "Untitled";
    $prolog{maxblanks} = -1;
    $prolog{order} = 'y,x';
}

sub storeVar()
{
    # the $follower is either empty or a dot or some blanks
    # if empty: ignore
    # if dot: ignore
    # if blanks:
    #    if $name ends with $ or '>' then add length(follower) to var display-length

    # xpos is only increased by  length($name)

    my ($ypos, $xpos, $name, $follower) = (@_);
    (!length($name)) and return ($ypos, $xpos);

    ($dbg =~/:ALL:|:SV:/) and print STDERR "\tSV: Y: $ypos, X: $xpos, N: [$name], F: [$follower]\n";
    
    # remove the leading dot
    $name =~ /^\.(.+)/ or die "\ninvalid varname: [$name]\n";
    my $baseName = $1;
    ++$xpos;

    my $screen_xpos = $xpos;    # save xpos

    my ($left, $right) = ($prolog{order} =~ /.*y.*x/) ? ($ypos, $xpos) : ($xpos, $ypos);
    ($prolog{order} =~ /^-/)  and $left = 9999 - $left ; 
    ($prolog{order} =~ /-.$/) and $right = 9999 - $right; 
    my $key = sprintf ("%05d:%05d", $left, $right);

    if (exists $yxParts{$key}) 
    {
        (!$dbg) and die "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";;
        print STDERR "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";
    }

    $xpos += length($baseName);

    # remove infinite '$'
    my $infinite;
    ($baseName =~ s/\$$//) and $infinite = 1;
    ($dbg =~/:ALL:|:SV:/ && $infinite) and print STDERR "\tSV: is infinite\n";
    # remove infinite '$'
    my $extend;
    ($baseName =~ s/>$//) and $extend = 1;
    ($dbg =~/:ALL:|:SV:/ && $extend) and print STDERR "\tSV: is extend\n";

    my $varLen = length($baseName);
    if (! ($infinite || $extend))
    {
        # remove trailing underscores
        $baseName =~ s/_*$//;
    }

    my $widgetWidth = $varLen+1;
    my %varAttribs= ( -y => $ypos , -x => $screen_xpos, -type => 'TextEntry' ,
                      -homeonblur      => 0, 
                      -bg              => 'white',
                      -showlines       => 1, 
                      -showoverflow    => 0, 
                      -wrapping        => 0,
                      name             => $baseName,
                      originalname     => $baseName);

    # TBC!
    if (!$infinite)
    {
        $varAttribs{-maxlength} = $varLen;
        $varAttribs{-width} = $varLen+1;        
    }
    else
    {
        (length($follower) && $follower ne '.') and $varAttribs{-width} = $varLen + length($follower) + 1;
        $varAttribs{-showoverflow} = 1;
    }
#    ($infinite) and $varAttribs{-width} = 0;

    if ($dbg=~/:ALL:|:SV:/)
    {
        foreach (keys %varAttribs)
        {
            print STDERR "\t\tSV a: $_ v: $varAttribs{$_}\n";
        }
    }

    $yxParts{$key} = { %varAttribs };
    return ($ypos, $xpos);
} # storeVar

sub storeTextViewer()
{
    # the $follower is either empty or a dot or some blanks
    # if empty: ignore
    # if dot: ignore
    # if blanks:
    #    if $name ends with $ or '>' then add length(follower) to var display-length

    # xpos is only increased by  length($name)

    my ($ypos, $xpos, $name, $follower) = (@_);
    ($dbg=~/:ALL:|:ST:/) and print STDERR "\tST: Y: $ypos, X: $xpos, N: [$name], F: [$follower]\n";
    
    # remove the leading dot
    $name =~ /^\.(.+)/ or die "\ninvalid varname: [$name]\n";
    my $baseName = $1;
    ++$xpos;
    my $screen_xpos = $xpos;    # save xpos
    my $key = sprintf ("%05d:%05d", $ypos, $xpos);

    if (exists $yxParts{$key}) 
    {
        (!$dbg) and die "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";;
        print STDERR "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";
    }

    $xpos += length($baseName);

    # remove infinite '$'
    my $infinite;
    ($baseName =~ s/\$$//) and $infinite = 1;

    # remove infinite '$'
    my $extend;
    ($baseName =~ s/>$//) and $extend = 1;

    my $varLen = length($baseName);

    if ($infinite || $extend)
    {
        if ($follower ne '.')
        {
            $varLen += length($follower);
        } 
    }
    else
    {
        # remove trailing underscores
        $baseName =~ s/_*$//;
    }

    my %varAttribs= ( -y => $ypos , -x => $screen_xpos, -type => 'TextViewer' ,
                      -maxlength        => $varLen, -bg=>'white',
                      -showoverflow     => 1, 
                      -maxlength        => $varLen,
                      -singleline       => 1,
                      -focusable        => 0,        
                      name             => $baseName,
                      originalname     => $baseName);

    # TBC!
    ($infinite) and $varAttribs{-maxlength} = 0;  

    if (!$infinite)
    {
        $varAttribs{-width} = $varLen;
    }

###    ($infinite && $follower eq "") and $tvAttribs{-width} = -1; # overwrite

    if ($dbg=~/:ALL:|:STV:/)
    {
        foreach (keys %varAttribs)
        {
            print STDERR "\t\tSTV a: $_ v: $varAttribs{$_}\n";
        }
    }

    $yxParts{$key} = { %varAttribs };
    return ($ypos, $xpos);
} # storeTextViewer

sub storeConst()
{
    my ($ypos, $xpos, $const, $follower, %userargs) = @_;
    (!length($const)) and return ($ypos, $xpos);

    (!%userargs) and %userargs = ();

    ($dbg=~/:ALL:|:SC:/) and print STDERR "\tSC: Y: $ypos, X: $xpos, C: [$const], F: [$follower]\n";

    my %constAttribs= ( -y => $ypos , -x => $xpos, -type => 'Label', %userargs);    
    my $key = sprintf ("%05d:%05d", $ypos, $xpos);

    if (exists $yxParts{$key}) 
    {
        (!$dbg) and die "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";;
        print STDERR "ERROR: unexpectedly detected a $key entry: [".$yxParts{$key}."]\n";
    }

    $const .= $follower;
    $xpos += length($const);

    # remove trailing blanks
    $const =~ s/\s*$//;

    $constAttribs{-text} = $const;
    $constAttribs{-width} = length($const);
    $constAttribs{name} = $const;
    $constAttribs{name} =~ s/[\W.:]/_/g; # replace any non-word character by underscore

    ($dbg=~/^:DCN:$/) and printf STDERR 
        "constant name at line %2d, column %2d, value: [$constAttribs{name}]\n", 
        $constAttribs{-y}, $constAttribs{-x};

    if ($dbg=~/:ALL:|:SC:/)
    {
        foreach (keys %constAttribs)
        {
            print STDERR "\t\tSC: a: $_ v: $constAttribs{$_}\n";
        }
    }
    $yxParts{$key} = { %constAttribs };
    return ($ypos, $xpos);
} # storeConst

# start reading


# read header
sub readProlog()
{
     # read STDIN until ^%% was found: the prolog
    while (<$inp>)
    {
        chomp;
        /^%%/ and last;
        /^\#/ and next;
        /^\s*$/ and next;
        (/^\s*(\w+)\s+(.*)/) and $prolog{lc($1)}=$2;
    }
    
    if (exists $prolog{rcfile})
    {
        $prolog{rcfilelabel} = &getText('F10');
    }
    else
    {
        $prolog{rcfilelabel} = "                ";
    }

    ($dbg=~/:ALL:|:RP:/) or return;

    # dump %prolog
    foreach (keys %prolog)
    {
        print STDERR "PROLOG, kwd: [$_], value: [$prolog{$_}]\n";
    }
} # readProlog

# read body
sub readBody()
{
    $ypos = -1;
    while (<$inp>)
    {
        chomp;
        /^%%/ and last;

        ++$ypos;

        my $xpos = $startx;;
        my $line = $_;
        my $entityLen = 0;
        my $const ="";
        my $follower = "";
        my $var = "";

        while (length($line))
        {
            my $oldLine = $line;

            # skip & eat up leading blanks
            if ($line =~ s/^(\s+)//)
            {
                my $ws = $1;
                ($dbg=~/:ALL:|:RBA:/) and printf STDERR "RBA1:ol[$oldLine],ws[$ws],c[$const],nl[$line]\n";

                if (!length($const))
                {
                    $xpos += length($ws);
                    next;
                }
                
                # we have a pending const
                if ($prolog{maxblanks} < 0)
                {
                    $const .= $ws;
                    ($dbg=~/:ALL:|:RBA:/) and printf STDERR "RBA2:ol[$oldLine],ws[$ws],c[$const],nl[$line]\n";
                    next;
                }
                                
                if (length($ws) <= $prolog{maxblanks})
                {
                    $const .= $ws;
                    ($dbg=~/:ALL:|:RBA:/) and printf STDERR "RBA3:ol[$oldLine],ws[$ws],c[$const],nl[$line]\n";
                    next;
                }
                
                # we have a pending const and too much blanks
                # -> write const and skip blanks
                ($dbg=~/:ALL:|:RBA:/) and printf STDERR "RBA4:ol[$oldLine],c[$const],nl[$line]\n";

                ($ypos, $xpos) = &storeConst($ypos, $xpos,  $const, "");
                $const="";

                $xpos += length($ws);
                next;
            }

            # read all non whitespace except '.'
            if ($line =~ s/^([^\s\.]*)(\.*)([[:alpha:]]*)/$2$3/x) # put last groups back
            {
                $const      .= $1;
                $follower    = $2;
                my $varstart = $3;
                ($dbg=~/:ALL:|:RBB:/) and printf STDERR "RBB1:ol[$oldLine],c[$const],f[$follower],vs[$varstart],nl[$line]\n";
                
                if (!length($varstart)) # no variable ahead
                {
                    $const .= $follower;
                    $line =~ s/^\.//x; # remove dot from input
                    ($dbg=~/:ALL:|:RBB:/) and printf STDERR "RBB2:ol[$oldLine],c[$const],f[$follower],nl[$line]\n";
                    next;
                }

                # a variable is ahead, flush $const
                ($ypos, $xpos) = &storeConst($ypos, $xpos,  $const, "");
                $const = "";
            }

            #==============================
            # this is the variable detector
            #==============================
            # read the variable (it's still in the input stream)
            # look for vars delimited by ' ' | '.' | $
            if ($line =~ s/^                              # from linestart
                           (\.[[:alpha:]]\w*>?\$*)(\ +|\.|$)  # $1 = var, $2 = what follows: blanks or a dot or EOL
                          /$2/x)                              # put the follower back, since it might be important
            {           
                $var = $1;
                $follower = $2;

                ($dbg=~/:ALL:|:RBC:/) and printf STDERR "RBC:ol[$oldLine],v[$var],f[$follower],nl[$line]\n";
                if (length($const)) # flush
                {
                    ($ypos, $xpos) = &storeConst($ypos, $xpos,  $const, "");
                    $const ="";
                }

                ($ypos, $xpos) = &storeVar($ypos, $xpos, $var, $follower);
                $var ="";
                next;
            }

            # if ($line =~ s/^(.?)+(\s*)$//) 
            # {
            #     my $const = $1;
            #   $follower = $2;
            #     ($dbg=~/:ALL:|:RBF:/) and printf STDERR "RBF:l[$oldLine],c[$const],f[$follower]\n";
            #     ($ypos, $xpos) = &storeConst($ypos, $xpos, $const, $follower);
            #   $const="";
            #     next;
            # }

            print STDERR "unregconized pattern, line ",$ypos+1,", line: [$oldLine]\n";
            last;
        } # while length($line)
        
        if (length($const))     # flush
        {
            ($ypos, $xpos) = &storeConst($ypos, $xpos, $const, "");
            $const="";
        }
    } # while (*inp)

    # write the bottomlines
    ($ypos < 20) and $ypos = 20;
    &storeConst($ypos, $startx, &getText('BOTTOMLINE1').$prolog{rcfilelabel}.&getText('BOTTOMLINE2'),"", -underline => '1');
    ++$ypos;
    # write the $_prompt-field
    &storeTextViewer($ypos, $startx-1, '._prompt$', "");

    (!$dbg=~/:ALL:|:RB:/) or return;

    # we have a hash of hashes here ...
    foreach (sort (keys %yxParts))
    {
        if (/(\d+):(\d+)/)
        {
            my ($y, $x) = ($1, $2);
            print STDERR "RB yxP $y:$x=\n";
            my $attribs = $yxParts{$_};

            foreach  (keys %$attribs)
            {
                print STDERR "\t\tRB a: $_, v: ", $attribs->{$_}, "\n";
            }
        }
    }
} # sub readBody()

# epilog
sub readEpilog()
{
    # read trailer
    while (<$inp>)
    {
        chomp;
        /^\#/ and next;
        # we may find field.attribute definitions here
        if (/^\s*([a-zA-Z]+[0-9a-zA-Z]*)\.(\S+)+\s+(.+?)\s*$/)  # hilighting comment
        {
            my $field = $1;
            my $attr = $2;
            my $val = $3;

            ($dbg=~/:ALL:|:FA:/) and print STDERR "FA: Field: [$field], Attr: [$attr]\ Value: [$val]\n" ;
            # get the field attribs
            (! exists $fieldAttribs{$field}) and $fieldAttribs{$field} = ();
            $fieldAttribs{$field}->{$attr} = &quoteIt($val); # ! this is a default setting. tune it, if appropriate
            $fieldAttribs{$field}->{originalname} = $field;
            if ($attr eq '-focusable')
            {
                $fieldAttribs{$field}->{-showlines} = $val;
            } elsif ($attr eq 'rename')
            {
                $fieldAttribs{$field}->{name} = $val;
            } elsif ($attr eq 'file' || $attr eq 'dir')
            {
                $needsCheckFD = 1;
                my $glob = 0;
                my ($min, $max) = (-1, -1);
                if ($val =~ /^(\d*):(\d*)/) # globbing
                {
                    $glob = 1;
                    $min = $1 ? $1 : 0;
                    $max = $2 ? $2 : -1;
                } elsif ($val =~ /^(\d+)/)
                {
                    $min = $max = $1;
                } else { printf STDERR "ERROR: field[$field], attribute[$attr]: can not parse file/directory count: [$val]\n"; }
                $fieldAttribs{$field}->{fdcheck} = $attr;
                $fieldAttribs{$field}->{glob} = $glob;
                $fieldAttribs{$field}->{min} = $min;
                $fieldAttribs{$field}->{max} = $max;
            } elsif ($attr eq 'default')
            {
                $fieldAttribs{$field}->{default} = "'$val'";
            } elsif ($attr eq 'composite')
            {
                $fieldAttribs{$field}->{composite} = "'$val'";
            }
        } elsif (/^\s*Label:([^\.]+)\.(\S+)+\s+(.+?)\s*$/i) 
        {
            # here we are at labels
            my $label = $1;
            my $attr = $2;
            my $val = $3;

            ($dbg=~/:ALL:|:LA:/) and print STDERR "LA: Label: [$label], Attr: [$attr]\ Value: [$val]\n" ;
            # get the label attribs
            (! exists $labelAttribs{$label}) and $labelAttribs{$label} = ();
            $labelAttribs{$label}->{$attr} = $val;
            $labelAttribs{$label}->{originalname} = $label;
#            ($attr eq 'rename') and $labelAttribs{$label}->{name} = $val;
            ($attr eq 'rename') and $labelAttribs{$val} = $labelAttribs{$label};
        }
    } # while *inp

    ($dbg =~ /:ALL:|:FA:|:LA:/) or return;

    # we have a hash of hashes here, dump them all  ...
    if ($dbg =~ /:ALL:|:FA:/)
    {
        foreach (sort (keys %fieldAttribs))
        {
            print STDERR "\tFA f: $_\n";
            my $attribs =$fieldAttribs{$_};
            foreach  (keys %$attribs)
            {
                print STDERR "\t\tFA a: $_, v: ", $attribs->{$_}, "\n";
            }
        }
    }

    if ($dbg =~ /:ALL:|:LA:/)
    {
        foreach (sort (keys %labelAttribs))
        {
            print STDERR "\tLA l: $_\n";
            my $attribs =$labelAttribs{$_};
            foreach  (keys %$attribs)
            {
                print STDERR "\t\tLA a: $_, v: ", $attribs->{$_}, "\n";
            }
        }
    }
}

sub mergeAttributes()
{
    foreach (keys %yxParts)
    {
        my $yxAttribs = $yxParts{$_};
        
        if ($yxAttribs->{-type } eq 'TextEntry' ||
            $yxAttribs->{-type } eq 'TextViewer')
        {
            my $yxName = $yxAttribs->{name};

            ($dbg =~ /:ALL:|:MA:/) and print STDERR  "MA, f: n: $yxName\n";
            if (exists $fieldAttribs{$yxName})
            {
                my $epilAtts = $fieldAttribs{$yxName};
                foreach my $epilAtt (keys %$epilAtts)
                {
                    ($dbg =~ /:ALL:|:MA:/) and print STDERR  "\tMA, f: eA: $epilAtt = ", $epilAtts->{$epilAtt}, "\n";
                    $yxAttribs->{$epilAtt} = $epilAtts->{$epilAtt};
                }
            }

            if ($dbg =~ /:ALL:|:MA:/) 
            {
                print STDERR "\tMA, f: new:\n";
                # dump the new values
                foreach my $att  (keys %$yxAttribs)
                {
                    print STDERR "\t\tMA, f: a: $att, v: ", $yxAttribs->{$att}, "\n";
                }
            }
        } 
        elsif ($yxAttribs->{-type } eq 'Label') # process labels
        {
            my $yxName = $yxAttribs->{name};

            ($dbg =~ /:ALL:|:MA:/) and print STDERR  "MA, l: n: $yxName\n";
            if (exists $labelAttribs{$yxName})
            {
                my $epilAtts = $labelAttribs{$yxName};
                foreach my $epilAtt (keys %$epilAtts)
                {
                    ($dbg =~ /:ALL:|:MA:/) and print STDERR  "\tMA, l: eA: $epilAtt = ", $epilAtts->{$epilAtt}, "\n";
                    $yxAttribs->{$epilAtt} = $epilAtts->{$epilAtt};
                }
            }

            if ($dbg =~ /:ALL:|:MA:/) 
            {
                print STDERR "\tMA, l: new:\n";
                # dump the new values
                foreach my $att  (keys %$yxAttribs)
                {
                    print STDERR "\t\tMA, l: a: $att, v: ", $yxAttribs->{$att}, "\n";
                }
            }
        }
    }
} # mergeAttributes

# write results
sub writeProlog()
{
    #----------------------------------- the script prolog
    print $out &getFragment('intro1');
    (exists $prolog{author})      and print $out &getFragment('header', type=>'AUTHOR     ', value=>$prolog{author});
    (exists $prolog{copyright})   and print $out &getFragment('header', type=>'COPYRIGHT  ', value=>$prolog{copyright});
    (exists $prolog{version})     and print $out &getFragment('header', type=>'VERSION    ', value=>$prolog{version});
    (exists $prolog{description}) and print $out &getFragment('header', type=>'DESCRIPTION', value=>$prolog{description});
    print $out &getFragment('intro2', PACKAGE=>$vslname);
    #----------------------------------- end of script prolog

    # create the %_fields hash
    foreach (sort (keys %ordYxParts))
    {   
        my $yxKey = $ordYxParts{$_}; # get the original key for yxParts
        my $field = $yxParts{$yxKey};

        if ($field->{-type} eq 'TextEntry' || 
            $field->{-type} eq 'TextViewer') # a var
        {
            if ($dbg =~ /:ALL:|:WSP:/)
            {
                print STDERR "\tWSP, Attribs:\n";
                print STDERR "\t\tWSP $_ = {", $field->{$_}, "}\n" foreach  (keys %$field);
            }

            my $name = $field->{name};
            (!$firstVar && $name !~ /^_/) and $firstVar = $name; # we need this later

            my $fieldDecl = &getFragment('field_decl', fieldname=>$name);
            my $indent = ' ' x length($fieldDecl);
            my $attHash = "value=>''";

            my $hasPrompt = 0;
            my $hasHelp = 0;

            foreach  (keys %$field)
            {
                my $att = $_;
                my $val = $field->{$att};
                                
                ($dbg =~ /:ALL:|:FDA:/) and print STDERR "FDA:n[$name],a[$att],v[$val]\n";

                ($att =~ /^-/) and next; # ignore dashed '-' attributes

                # some parts are printed at the end
                ($att =~ /^help$/)   and $hasHelp   = 1 and next;
                ($att =~ /^prompt$/) and $hasPrompt = 1 and next;
                
                $attHash .= &getFragment('field_attr', attr=>$att, attrval=>&quoteIt($val));
            }
            
            ($hasPrompt) and $attHash .= &getFragment('field_attr', attr=>'prompt', attrval=>&quoteIt($field->{prompt}));
            ($hasHelp)   and $attHash .= &getFragment('field_attr', attr=>'help', attrval=>&quoteIt($field->{help}));

            $attHash .= &getFragment('field_decl_close', fieldname => $name);
            ($dbg =~ /:ALL:|:AHA:/) and print STDERR "AHA:h[$attHash]\n";

            $Text::Wrap::break = '@@'; # only break at @@
            my $outstring = wrap($fieldDecl, $indent, ($attHash));
            $outstring =~ s/@@/ /g;
            print  $out $outstring;
        } # TextEntry or TextViewer
    } # foreach (sort (keys %ordYxParts))

    print $out &getFragment('fields_hash_close');
    (exists $prolog{rcfile}) and print $out &getFragment('rcfile_decl', rcfile=>$prolog{rcfile});

    # create some onXxx and Help, checkParms, setValues funcs
    print $out &getFragment('helper_funcs', FIELD=>&getText('FIELD'), FIELDNAME=>&getText('FIELDNAME'),
                       NOHELP=>&getText('NOHELP'),HELP=>&getText('HELP'), 
                       UNKNOWNPFMT=>&getText('UNKNOWNPFMT'), INVALIDPARM=>&getText('INVALIDPARM'));
    
    (exists $prolog{rcfile}) and print $out &getFragment('rcfile_rw', rcfile=>$prolog{rcfile},                                                   
                                                    COULDNTREADRC=>&getText('COULDNTREADRC', RCFILE=>'$_rcFile'),
                                                    COULDNTOPENRC=>&getText('COULDNTOPENRC', RCFILE=>'$_rcFile'),
                                                    VALUESSAVED=>&getText('VALUESSAVED'));
} # writeScriptProlog()

sub writeDialog()
{
    #----------------------------------- mkdialog - intro
    print $out &getFragment('mkdialog_intro', ptitle => $prolog{title});
    &writeEntities();
    print $out &getFragment('mkdialog_bindings', firstvar => $firstVar);
    (exists $prolog{rcfile}) and print $out &getFragment('key_f10_binding');    
    print $out &getFragment('close_function', function => 'mkdialog');
} # writeDialog

sub orderEntities()
{
    # we internallly us a copy of %xyParts because we have to regard 
    # the order which has possibly been redefined by the user

    foreach my $key (sort (keys %yxParts))
    {
        my $yxAttr = $yxParts{$key};
        my $order = (exists $yxAttr->{order}) ? $yxAttr->{order} : 0;
        my $newKey = sprintf ("%03d:%s", $order, $key);
        ($dbg =~ /:ALL:|:ORD:/) and print STDERR  "ORD:ok[$key],nk[$newKey]\n";
        $ordYxParts{$newKey} =  $key ;
    }
}

sub writeEntities()
{
    foreach (sort (keys %ordYxParts))
    {   
        my $yxKey = $ordYxParts{$_}; # get the original key for yxParts
        my $yxAttr = $yxParts{$yxKey};
        my $statement = "# ?".$yxAttr->{-type}."\n";
        if ($yxAttr->{-type} eq 'TextEntry') # a var
        {
            my $originalName = $yxAttr->{originalname};
            my $name = $yxAttr->{name};

            $yxAttr->{-onfocus} = '\&on_'.${name}.'Focus';
            $yxAttr->{-onchange} = '\&on_'.${name}.'Change';
            $yxAttr->{-onblur} = '\&on_'.${name}.'Blur';
            $yxAttr->{-text} =    "\$_fields{$name}->{value}";
            $statement = sprintf ('$_mainWin->add(\'%s\', \'TextEntry\'', $name); 
        } elsif ($yxAttr->{-type} eq 'Label') # a label
        {
            $statement = sprintf('$_mainWin->add(\'y%02dx%03dLabel\', \'Label\'', $yxAttr->{-y},$yxAttr->{-x});
        } elsif ($yxAttr->{-type} eq 'TextViewer') # a TextViewer
        {
            my $name = $yxAttr->{name};
            my %entAttribs = ( -y        => $yxAttr->{-y},
                               -x        => $yxAttr->{-x}, 
                               -width    => $yxAttr->{-width},
                               -name     => '$varname',
                               -text     => ''); #\$${name});
                               
            $statement = sprintf ('$_mainWin->add(\'%s\', \'TextViewer\'', $name); 
        }

        # here, we have a string fragment which has to to completed.
        $statement =~ s/\n$//;  # delete newline
        my $indent = index($statement, "(") + 1;
        my $blanks = ' ' x $indent;
        my $linelen = length($statement) + $indent;

        # first, print x and y
        $statement .= ', -y=>'.$yxAttr->{-y};
        $statement .= ', -x=>'.$yxAttr->{-x};

        my $hasText = 0;

        foreach  (keys %$yxAttr)
        {
            my $att = $_;
            
            ($att !~ /^-/ && $att ne 'name') and next; # only process dashed '-' attributes and 'name'
            my $val = $yxAttr->{$att};

            ($dbg =~ /:ALL:|:XYA:/) and print STDERR "\tXYA:name[$yxAttr->{name}],att[$att],val[$val]\n";
            ($att =~ /^-type$/)  and next; # internal use only

            ($dbg =~ /:ALL:|:WEN:/) and print STDERR "\tWEN, statement: $att = $val\n";

            # some parts are printed at the end
            ($att =~ /^-text$/)  and $hasText   = 1 and next;
            ($att =~ /^-y/) and next;
            ($att =~ /^-x/) and next;
            
            $statement .= ", ";
            if ($linelen > 72)
            {
                $statement .="\n    ".$blanks;
                $linelen = 4 + $indent; # 
            }
            $val = &quoteIt($val);
            $statement .= $att."=>".$val;
            $linelen += length($att."=>".$val);
        }

        ($hasText)   and $statement .= ",\n    ${blanks}-text=>".&quoteIt($yxAttr->{-text});

        print $out "    ", $statement, ");\n";
    }    
}# writeEntities()

sub quoteIt()
{
    my $val = shift;
    # (!$val) and return undef;
    
    # ($dbg =~ /:ALL:|:QUO:/) and print STDERR "QUO:v[$val],r[".ref($val)."]\n";
    # (ref($val) eq 'REF') and return $val;

    #    From the "Progrmming Perl" book by Wall, Christiansen & Schwartz, published by O'Reilly's
    ($val =~ /^(\d+\.?\d*|\.\d+)$/) and return $val;  # match valid number (isNumeric)
    ($val =~ /^[\'\"\\\$]/) and return $val; # already quoted or var/ref ...
    return  "'".$val."'";
}

sub writeOnFunctions()
{
    #----------------------------------- the "on" functiuons
    foreach (sort (keys %yxParts))
    {
        my $attribs = $yxParts{$_};
        ($attribs->{-type} ne 'TextEntry') and next; # not a var
        print $out &getFragment('on_funcs', '.fname.'=>$attribs->{name}, hname=>$attribs->{name});
    } # foreach ... 
    print $out &getFragment('on_funcs_end');
} # writeOnFunctions

sub writeEpilog()
{
    my $writeRcCall = "";
    (exists $prolog{rcfile}) and $writeRcCall = '&WriteRcfile();';
    print $out &getFragment('quit_func');
    ($needsCheckFD) and print $out &getFragment('checkfd_func', 
                                           FIELD_AND_PATTERN => &getText('FIELD_AND_PATTERN'),
                                           NOTEXISTHASTO=>&getText('NOTEXISTHASTO'),
                                           FATALCHECKCOND=>&getText('FATALCHECKCOND'),
                                           NOSUCHFILE=>&getText('NOSUCHFILE'),
                                           TYPEEXISTS=>&getText('TYPEEXISTS'),
                                           COUNTTYPEEXIST=>&getText('COUNTTYPEEXIST'),
                                           MORETHANEXIST=>&getText('MORETHANEXIST'),
                                           LESSTHANEXIST=>&getText('LESSTHANEXIST'),
                                           FATALCHECKCOND=>&getText('FATALCHECKCOND'));
    print $out &getFragment('check_field', MANDATORY=>&getText('MANDATORY'));
    print $out &getFragment('check_screen_intro1');
    print $out &getFragment('check_screen_intro3');

    foreach (sort (keys %ordYxParts)) # use sequence by 'order' attribute
    {   
        my $yxKey = $ordYxParts{$_}; # get the original key for yxParts
        my $yxAttr = $yxParts{$yxKey};

        my $name = $yxAttr->{name};
        (!$name) and next;

        ($yxAttr->{-type} ne 'TextEntry') and next;     
        print $out &getFragment('dofieldcheck', fieldname => $name);
    }

    print $out &getFragment('runepilog1',
                       PROCINPROGRESS=>&getText('PROCINPROGRESS'),
                       DONE=>&getText('DONE'),
                       LOG=>&getText('LOG'));

    (exists $prolog{rcfile}) and print $out &getFragment('writercfile_call');

    print $out &getFragment('runepilog2',
                       PROCINPROGRESS=>&getText('PROCINPROGRESS'),
                       DONE=>&getText('DONE'),
                       LOG=>&getText('LOG'));

    (exists $prolog{rcfile}) and print $out &getFragment('readrcfile_call');
    print $out &getFragment('final');
} # writeEpilog

sub Usage()
{
    system "perldoc $0";
    exit;
}

sub Version() 
{ 
    print STDERR <<EOF;

crtvslcmd.pl $vsn

Copyright (C) 2010 Edgar <emax> Hermanns.

This is free software; see the source for copying conditions.  
There is NO warranty; not even for MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE.

EOF
exit; 
};

sub writePkg()
{
    (-f "$vslpath") and return;
    open (PKG, "> $vslpath") or die ("\n\ncouldn't create package file [$vslpath]\n\n");
    print PKG &getFragment('pkg', PKG=>$vslname, 
			   VALUESENTERED=>&getText('VALUESENTERED'),
                           PARMSNVALUES=>&getText('PARMSNVALUES'),
                           PARMSNVALUESUL=>&getText('PARMSNVALUESUL'));
    close PKG;
}

# ------------------------------ main


GetOptions  ('help|?|h'          => \$helpme, 
	     'version|v'         => \$version,
             'lang=s'            => \$lang,
             'input=s'           => \$input,
             'output=s'          => \$output,
             'debug=s'           => \$dbg);

($helpme) and  &Usage();
($version) and &Version(); 

if (!exists $texts{$lang})
{
    print STDERR "WARNING: language [$lang] is currently not supported, switching to [de]\n";
    $lang = 'de';
}

(!$input) and die ("\n\n$0: missing --input parameter\n\n");
(!$output) and die ("\n\n$0: missing --output parameter\n\n");
(! -f $input) and die ("\ninput file [$input] not found.\n\n");
open ($inp, "< $input") or die ("\ncouldn't open input file [$input]\n\n");
open ($out, "> $output") or die ("\ncouldn't open output file [$output]\n\n");

$vslname = basename($input, ".vsl");
$vslpath=dirname($output)."/$vslname.pm";

&setDefaults();

# start reading
&readProlog();
&readBody();
&readEpilog();

close $inp;

# merge what we found in epilog with entities
&mergeAttributes();

# reorder entities due to 'order' attribute
&orderEntities();

# write results
&writeProlog();
&writeDialog();
&writeOnFunctions();
&writeEpilog();
&writePkg();

close $out;
